<!---
	This tag can create anonymous functions inside other functions with compiler magic.
	
	Example:
	
	<cffunction name="test" access="public" returntype="any">
	
		<cfset var doSomething = "">
	
		<cc:function name="doSomething" arguments="a">
			var i = 0;
			for( i=1; i lte 10; i++ )
				writeOutput("Hello World##arguments.a##<br>");
		</cc:function>
	
		<cfset doSomething(1)>
	</cffunction>
--->
<cfif thisTag.executionMode eq "end">

	<cfscript>
		// NOTE:
		// Unless there's at least one function then appending to the metadata
		// collection causes a divide by zero exception.
		function makeMetaDataWork() {}
	</cfscript>

	<!--- Required parameters --->
	<cfparam name="attributes.arguments" type="string" default="">
	<cfparam name="attributes.name" type="string">

	<!--- Use the metadata from this page for static caching --->
	<cfset metadata = getMetaData(getPageContext().getPage())>

	<!--- Code source is the generated content --->
	<cfset source = thisTag.generatedContent>
	
	<!--- Don't output the source! --->
	<cfset thisTag.generatedContent = "">

	<!--- Create the cache of compiled functions --->
	<cfif not structKeyExists(metadata,"compiledFunctions")>
		<cflock name="FunctionCompilerLock" timeout="10">
			<cfif not structKeyExists(metadata,"compiledFunctions")>
				<cfset metadata.compiledFunctions = structNew()>
			</cfif>
		</cflock>
	</cfif>
		
	<!--- Wrap the function in the script func. --->
	<cfset source = "<cfscript>function func(#attributes.arguments#) { #source# }</cfscript>">
	
	<!--- Get the compiled function cache --->
	<cfset compiledFunctions = metadata.compiledFunctions>
	
	<!--- Check of the function exists in the cache --->
	<cfif not structKeyExists(compiledFunctions,source)>
		
		<!---
			This next block of code is a mess that needs documenting. It essentially invokes
			the CF compiler on a string of source code and gets the compiled class back.
			
			It then takes the function off the compiled class and sets it to the name in the caller.
		--->
		
		<cfset ServletContext = getPageContext().getServletContext().getContext("/")>
		<cfset c = createObject("java","coldfusion.compiler.NeoTranslator").init(ServletContext)>
		<cfset cftags = c.getClass().getDeclaredField("cftags")>
		<cfset sourceFile = createObject("java","java.io.File").init("fake")>
		<cfset translationContext = createObject("java","coldfusion.compiler.NeoTranslationContext").getClass().getDeclaredConstructors()>
		<cfset args = arrayNew(1)>
	
		<cfset cftags.setAccessible(true)>
		<cfset cftags = cftags.get(c)>

		<cfset args[1] = cftags>
		<cfset args[2] = ServletContext>
		<cfset args[3] = sourceFile>
		<cfset args[4] = "compiledClass">
		<cfset translationContext[1].setAccessible(true)>
		<cfset TranslationContext = translationContext[1].newInstance(args)>

		<cfset stream = createObject("java","coldfusion.compiler.StringCharStream").init(source)>
		<cfset compiler = createObject("java","coldfusion.compiler.cfml40").init(stream)>
		<cfset compiler.setTranslationContext(TranslationContext)>
		<cfset result = compiler.start()>

		<cfset assembler = createObject("java","coldfusion.compiler.TemplateAssembler").getClass().getDeclaredConstructors()>
		<cfset assembler = assembler[1]>
		<cfset assembler.setAccessible(true)>
		<cfset assembler = assembler.newInstance(arrayNew(1))>

		<cfset assemble = assembler.getClass().getDeclaredMethods()>
		<cfset assemble = assemble[2]>
		<cfset assemble.setAccessible(true)>

		<cfset context = TranslationContext>

		<cfset args = [result.getClass()]>
		<cfset validate = c.getClass().getDeclaredMethod("validate", args)>
		<cfset validate.setAccessible(true)>
		<cfset args[1] = result>
		<cfset validate.invoke(c,args)>

		<!---
			This is a black magic combination of compiler magic that makes the assembler useful.
		--->
		<cfset Treewalker = createObject("java","coldfusion.compiler.Treewalker")>
		<cfset SemanticAnalyzer = createObject("java","coldfusion.compiler.SemanticAnalyzer")>
		<cfset SemanticAnalyzer = SemanticAnalyzer.getClass().getDeclaredConstructors()>
		<cfset SemanticAnalyzer = SemanticAnalyzer[1]>
		<cfset SemanticAnalyzer.setAccessible(true)>
		<cfset args = [context]>
		<cfset SemanticAnalyzer = SemanticAnalyzer.newInstance(args)>
		<cfset ExprVisitor = createObject("java","coldfusion.compiler.ExprVisitor").init(context)>
		<cfset TemplateOptimizer = createObject("java","coldfusion.compiler.TemplateOptimizer")>
		<cfset TemplateOptimizer = TemplateOptimizer.getClass().getDeclaredConstructors()>
		<cfset TemplateOptimizer = TemplateOptimizer[1]>
		<cfset TemplateOptimizer.setAccessible(true)>
		<cfset TemplateOptimizer = TemplateOptimizer.newInstance(arrayNew(1))>

		<cfset Treewalker.postorder(result,SemanticAnalyzer)>
		<cfset Treewalker.postorder(result,ExprVisitor)>
		<cfset Treewalker.postorder(result,TemplateOptimizer)>

		<cfset args = [result,context]>
		
		<!---
			This makes the class loader happy.
		--->
		<cfset result = assemble.invoke(assembler,args.toArray())>
		<cfset JURL = createObject("java","java.net.URL").init("http://localhost/cfide/dynamic-function")>
		<cfset CodeSource = 
			createObject("java","java.security.CodeSource")
				.init(JURL,javaCast("java.security.cert.Certificate[]",arrayNew(1)))>

		<cfset TemplateClassLoader =
		 	createObject("java","coldfusion.runtime.TemplateClassLoader")
				.init(result,CodeSource,"")>

		<!--- 
			Now get the class back from the class loader and try to make it all work.
		--->
		<cfset args = [createObject("java","java.lang.String").getClass()]>
		<cfset findClass = TemplateClassLoader.getClass().getDeclaredMethod("findClass",args)>
		<cfset findClass.setAccessible(true)>
		<cfset args[1] = "compiledClass">
		<cfset compiledClass = findClass.invoke(TemplateClassLoader,args).newInstance()>
		
		<!--- This stuff is probably not needed since we just want the function off it. --->
		<cfset pageContext = compiledClass.initialize(getPageContext(),javaCast("null",""))>
		<cfset pageContext.propagateDefaultQueryScope(getPageContext())>
		
		<cfset compiledClass.invoke(getPageContext().getFusionContext())>

		<!--- Now get the function off the class --->
		<cfset func = pageContext.getPage().getClass().getDeclaredField("func")>
		<cfset func.setAccessible(true)>
		<cfset func = func.get(pageContext.getPage())>
		
		<!--- Set the function into the caller --->
		<cfset metadata.compiledFunctions[source] = func>
					
	<cfelse>
		<!--- Grab the function out of the cache if it exists --->
		<cfset func = metadata.compiledFunctions[source]>
	</cfif>
	
	<!--- Set the function into the caller scope. --->
	<cfset caller[attributes.name] = func>
</cfif>