<cfcomponent>

	<cfset Class = createObject( "java", "java.lang.Class")>
	<cfset UDFMethod = Class.forName("coldfusion.runtime.UDFMethod")>
	<cfset LocalScope = Class.forName("coldfusion.runtime.LocalScope")>
	<cfset Object = Class.forName("java.lang.Object")>
	<cfset CFPage = Class.forName("coldfusion.runtime.CFPage")>
	<cfset ArgumentCollection = Class.forName("coldfusion.runtime.ArgumentCollection")>
	<cfset TemplateProxy = Class.forName("coldfusion.runtime.TemplateProxy")>
	
	<!--- Must use createObject() to get the right classloader --->
	<cfset ServiceFactory = createObject("java","coldfusion.server.ServiceFactory")>
	<cfset RuntimeService = ServiceFactory.getRuntimeService()>
	<cfset DatasourceService = ServiceFactory.getDatasourceService()>
	<cfset Array = createObject("java","java.lang.reflect.Array")>	
	
	<cfset JavaMethods = structNew()>

	<cfset initJavaMethods()>

	<cffunction name="init" access="public" returntype="any" output="false">
			
		<cfreturn this>
	</cffunction>
		
	<cffunction name="initJavaMethods" access="private" returntype="void" output="false">
	
		<cfset var args = arrayNew(1)>
		
		<cfset args[1] = LocalScope>
		<cfset args[2] = Object>
		<cfset args[3] = CFPage>
		<cfset args[4] = ArgumentCollection>

		<cfset JavaMethods.UDFMethod.runFunction = UDFMethod.getDeclaredMethod("runFunction",args)>
	</cffunction>
	
	<cffunction name="getApplicationCFC" access="public" returntype="any" output="false">
	
		<cfset var result = application.getEventInvoker().getClass().getSuperClass().getDeclaredField("template")>
		<cfset result.setAccessible(true)>
		<cfreturn result.get(application.getEventInvoker())>
	</cffunction>
	
	<cffunction name="runFunction" access="public" returntype="struct" output="false">
		<cfargument name="context" type="any" required="true">
		<cfargument name="function" type="any" required="true">
		<cfargument name="funcArgs" type="struct" required="false" default="#structNew()#">
	
		<cfset var args = arrayNew(1)>
		<cfset var field = "">
		<cfset var result = structNew()>
		<cfset var i = "">
		<cfset var argList = "">
	
		<!--- The localScope is the same as the var'ed variables "scope" in functions --->
		<cfset args[1] = createObject("java","coldfusion.runtime.LocalScope").init()>
		
		<!---
			Figure out what the target context of this method call is. Is it a 
			cfc instance or a page?
		--->
		<cfif isObject(arguments.context) and getMetaData(arguments.context).getClass().getName() eq "coldfusion.runtime.AttributeCollection">
			<cfset args[2] = arguments.context>
			<cfset field = TemplateProxy.getDeclaredField("page")>
			<cfset field.setAccessible(true)>
			<cfset args[3] = field.get(arguments.context)>
		<cfelseif isObject(arguments.context) and arguments.context.getClass().getName() eq "coldfusion.runtime.NeoPageContext">
			<cfset args[3] = arguments.context.getPage()>
		<cfelse>
			<cfthrow type="InvalidContext" message="No valid context provided for function call.">			
		</cfif>
		
		<!--- 
			Allow passing a string instead of a function pointer for resolving 
			the function name.
		--->
		<cfif isSimpleValue(arguments.function)>
			<cfset arguments.function = args[3][arguments.function]>
		</cfif>
		
		<!--- 
			Ensure the arguments collection is synched up so there's a numeric 
			index for each parameter and that it matches the named one 
			so that the below example conflict is impossible.
			
			Internally CF uses the number index first so we do too. Also ensures 
			the existance of both vars. For example without this code you get 
			nasty errors if you use arguments.arg and only arg[1] was passed.
			
			Ex.
			
			function foo(arg) {}
			
			args[1] = 10;
			args.arg = 20;
		--->
		<cfset argList = getMetaData(arguments.function).parameters>
		<cfloop from="1" to="#arrayLen(argList)#" index="i">
			<cfset field = argList[i].name>
			<cfif not structKeyExists( arguments.funcArgs, i ) and structKeyExists( arguments.funcArgs, field )>
				<cfset arguments.funcArgs[i] = arguments.funcArgs[field]>
				<cfset arguments.funcArgs[field] = arguments.funcArgs[i]>
			</cfif>
			<cfif (structKeyExists( arguments.funcArgs, i ) and not structKeyExists( arguments.funcArgs, field )) or (arguments.funcArgs[i] neq arguments.funcArgs[field])>
				<cfset arguments.funcArgs[field] = arguments.funcArgs[i]>
				<cfset arguments.funcArgs[i] = arguments.funcArgs[field]>
			</cfif>
		</cfloop>
		
		<cfset args[4] = createObject("java","coldfusion.runtime.ArgumentCollection").init(arrayNew(1),arguments.funcArgs)>
		
		<cfset result.returnValue = JavaMethods.UDFMethod.runFunction.invoke( arguments.function, args )>
		<cfset result.localScope = args[1]>
			
		<cfif not isDefined("result.returnValue")>
			<cfset result.returnValue = "">
		</cfif>
			
		<cfreturn result>
	</cffunction>
	
	<!--- oddly enough this is VERY expensive compared to evaluate() --->
	<cffunction name="invokeJava" access="public" returntype="any" output="false">
		<cfargument name="instance" type="any" required="true">
		<cfargument name="methodName" type="string" required="true">
		<cfargument name="args" type="array" required="true">
		
		<cfset var ObjectHandler = createObject("java","coldfusion.runtime.java.ObjectHandler")>
		<cfset var handler = ObjectHandler.init(getMetaData(arguments.instance))>
		<cfset var method = handler.findMethod(arguments.methodName,arguments.args)>
		<cfreturn method.invoke(arguments.instance,arguments.args)>
	</cffunction>	
		
	<cffunction name="fixUrlParameters" access="public" returntype="void" output="false">
		<cfset var params = createObject("java","coldfusion.util.HTMLTools").parseQueryString(cgi.query_string)>
		<cfset structAppend( url, params, true )>
	</cffunction>

	<cffunction name="getMappings" access="public" returntype="any" output="false">
		<cfreturn RuntimeService.getMappings()>
	</cffunction>
		
	<cffunction name="getCustomTagPaths" access="public" returntype="any" output="false">	
		<cfreturn RuntimeService.getCustomTags()>
	</cffunction>

	<cffunction name="addCustomTagPath" access="public" returntype="boolean" output="false">		
		<cfargument name="path" type="string" required="true">

		<cfset var paths = RuntimeService.getCustomTags()>
		<cfif not structKeyExists(paths,path) >
			<cfset paths[path] = path>
			<cfreturn true>
		</cfif>
		<cfreturn false>
	</cffunction>
	
	<cffunction name="removeCustomTagPath" access="public" returntype="numeric" output="false">
		<cfset var paths = RuntimeService.getCustomTags()>
		<cfset var item = 0>
		<cfset var ret = 0>
		<cfloop collection="#paths#" item="item">
			<cfif item eq path>
				<cfset structDelete( paths, item )>
				<cfset ret = ret + 1>
			</cfif>
		</cfloop>
		<cfreturn ret>
	</cffunction>
	
	<cffunction name="getRequestSettings" access="public" returntype="any" output="false">
		<cfreturn duplicate(RuntimeService.getRequestSettings())>
	</cffunction>
	
	<cffunction name="setLogSlowRequests" access="public" returntype="void" output="false">		
		<cfargument name="value" type="boolean" required="true">
	
		<cfset var settings = RuntimeService.getRequestSettings()>
		<cfset settings["logslowrequests"] = value>
	</cffunction>
	
	<cffunction name="setSlowRequestTimeout" access="public" returntype="void" output="false">		
		<cfargument name="value" type="numeric" required="true">
	
		<cfset var settings = RuntimeService.getRequestSettings()>
		<cfset settings["slowrequesttimeout"] = value>
	</cffunction>
	
	<cffunction name="setRequestTimeoutLimit" access="public" returntype="void" output="false">		
		<cfargument name="value" type="numeric" required="true">
	
		<cfset var settings = RuntimeService.getRequestSettings()>
		<cfset settings["timeoutRequestTimeLimit"] = value>
	</cffunction>	

	<cffunction name="setTimeoutRequests" access="public" returntype="void" output="false">		
		<cfargument name="value" type="boolean" required="true">
	
		<cfset var settings = RuntimeService.getRequestSettings()>
		<cfset settings["timeoutRequests"] = value>
	</cffunction>	
	
	<!--- do these actually work? --->
	<cffunction name="setSuppressWhitespace" access="public" returntype="void" output="false">
		<cfargument name="value" type="boolean" required="true">	

		<cfset RuntimeService.setWhitespace( value )>
	</cffunction>
	
	<cffunction name="isSuppressingWhitespace" access="public" returntype="boolean" output="false">
		<cfreturn RuntimeService.getWhitespace()>
	</cffunction>
	
	<cffunction name="setSaveClassFiles" access="public" returntype="void" output="false">
		<cfargument name="value" type="boolean" required="true">	
	
		<cfset RuntimeService.setSaveClassFiles( value )>
	</cffunction>

	<cffunction name="isSavingClassFiles" access="public" returntype="boolean" output="false">	
		<cfreturn RuntimeService.getSaveClassFiles()>
	</cffunction>
	
	<!--- lots more things like this in RuntimeService --->
	
	<cffunction name="getTablesForDatasource" access="public" returntype="any" output="true">
		<cfargument name="datasource" type="string" required="true">
		<cfargument name="tableTypes" type="string" required="false">	
		<cfargument name="tableCatalog" type="string" required="false">
		
		<cfset var dsm = DatasourceService.getDatasource(datasource).getConnection().getMetaData()>
		<cfset var methods = dsm.getClass().getDeclaredMethods()>
		<cfset var method = 0>
		<cfset var getTables = 0>
		<cfset var i = 0>
		<cfset var args = arraynew(1)>
		<cfset var tables = 0>
		<cfset var resultMeta = 0>
		<cfset var result = 0>
		<cfset var cols = "">
		<cfset var tableFilter = 0>

		<cfset arrayResize( args, 4 )>
				
		<cfif structKeyExists(arguments,"tableCatalog")>
			<cfset args[1] = arguments.tableCatalog>
		</cfif>
		
	
		<cfif structKeyExists(arguments,"tableTypes")>
			<cfset tableFilter = array.newInstance(class.forName("java.lang.String"),listLen(arguments.tableTypes))>
			<cfloop from="1" to="#listLen(arguments.tableTypes)#" index="i">
				<cfset Array.set(tableFilter,javaCast("int",i-1),ucase(listGetAt(arguments.tableTypes,i)))>
			</cfloop>
			<cfset args[4] = tableFilter>
		</cfif>
			
		<!--- Cheap way around bad CF method resolution :( --->
		<cfset method = methods[120]>
		<cfif method.getName() neq "getTables">
			<cfloop from="1" to="#arraylen(methods)#" index="i">
				<cfif methods[i].getName() eq "getTables">
					<cfset method = methods[i]>
					<cfbreak>
				</cfif>
			</cfloop>
		</cfif>
		
		<cfset method.setAccessible(true)>
		
		<cfset tables = method.invoke(dsm,args)>
		<cfset resultMeta = tables.getMetaData()>

		<cfloop from="1" to="#resultMeta.getColumnCount()#" index="i">
			<cfset cols = listAppend(cols,resultMeta.getColumnName(javaCast("int",i)))>
		</cfloop>
		
		<cfset result = queryNew(cols)>
		<cfset cols = listToArray(cols)>
	
		<cfloop condition="#tables.next()#">
			<cfset queryAddRow(result)>
			<cfloop from="1" to="#arrayLen(cols)#" index="i">
				<cfset querySetCell(result,cols[i],tables.getString(cols[i]))>
			</cfloop>
		</cfloop>
		
		<cfset tables.close()>
		
		<cfreturn result>
	</cffunction>
	
</cfcomponent>